<?php

namespace App\Http\Controllers;

use Illuminate\Http\Request;
use Illuminate\Http\RedirectResponse;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Redirect;
use Illuminate\Support\Facades\Storage;
use App\Models\User;

class ProfileController extends Controller
{
    /**
     * Update the user's profile information, including profile picture.
     */
    public function update(Request $request): RedirectResponse
    {
        $validated = $request->validate([
            'full_name' => 'required|string|max:255',
            'designation' => 'nullable|string|max:255',
            'contact_number' => 'nullable|string|max:255',
            // 'email' => 'required|email|max:255|unique:users,email,' . auth()->id(),
            'department' => 'nullable|string|max:255',
            'profile_picture' => 'nullable|image|mimes:jpeg,png,jpg,gif|max:2048',
        ]);

        $user = Auth::user();

        // Handle profile picture update
        if ($request->hasFile('profile_picture')) {
            // Delete the old profile picture if it exists
            if ($user->profile_picture) {
                Storage::disk('public')->delete($user->profile_picture);
            }

            // Store the new file
            $validated['profile_picture'] = $request->file('profile_picture')->store('profile_pictures', 'public');
        }

        // Update user details
        $user->update($validated);

        return back()->with('success', 'Profile updated successfully!');
    }

    /**
     * Delete the user's account.
     */
    public function destroy(Request $request): RedirectResponse
    {
        $request->validateWithBag('userDeletion', [
            'password' => ['required', 'current_password'],
        ]);

        $user = $request->user();

        // Delete the user's profile picture if it exists
        if ($user->profile_picture) {
            Storage::disk('public')->delete($user->profile_picture);
        }

        Auth::logout();

        $user->delete();

        $request->session()->invalidate();
        $request->session()->regenerateToken();

        return Redirect::to('/');
    }

    public function guest($id)
    {
        $viewUser = User::findOrFail($id);
        // dd($user);

        $certificates = $viewUser->certificates()->where('shared', true)->get();
        return view('certificates.guest', compact('viewUser', 'certificates'));
    }

    /**
     * Show the authenticated user profile page.
     */
    public function user($id)
    {
        $viewUser = User::findOrFail($id);
        // dd($user);
        $certificates = $viewUser->certificates()->where('shared', true)->get();
        return view('certificates.auth', compact('viewUser', 'certificates'));
    }

    /**
     * Submit a rating for a user.
     */
    public function rate(Request $request, $userId)
    {
        // Validate the incoming request
        $request->validate([
            'rating' => 'required|integer|min:1|max:5',
        ]);

        // Find the user to be rated
        $user = User::findOrFail($userId);

        // Create a new review entry
        Review::create([
            'user_id' => $user->id,          // The user being reviewed
            'reviewer_id' => auth()->id(),   // The logged-in user giving the review
            'rating' => $request->rating,    // The rating submitted
        ]);

        return redirect()->back()->with('success', 'Thank you for your rating!');
    }
    /**
     * Update the role of a user.
     */
    public function updateRole(Request $request)
    {
        $request->validate([
            'user_id' => 'required|exists:users,id',
            'role' => 'required|in:Employee,Supervisor',
        ]);

        $user = User::findOrFail($request->user_id);
        $user->role = $request->role;
        $user->save();

        return back()->with('success', 'User role updated successfully.');
    }

    /**
     * Update the status of a user.
     */
    public function updateStatus(Request $request)
    {
        $request->validate([
            'user_id' => 'required|exists:users,id',
            'status' => 'required|in:Active,Deactive',
        ]);

        $user = User::findOrFail($request->user_id);
        $user->status = $request->status;
        $user->save();

        return back()->with('success', 'User status updated successfully.');
    }

    public function adminUpdateUserRole(Request $request)
    {
        $request->validate([
            'user_id' => 'required|exists:users,id',
            'role' => 'required|in:Employee,Supervisor',
        ]);

        $user = User::findOrFail($request->user_id);
        $user->role = $request->role;
        $user->save();

        return response()->json(['success' => true, 'message' => 'User role updated successfully.']);
    }

    public function adminUpdateUserStatus(Request $request)
    {
        $request->validate([
            'user_id' => 'required|exists:users,id',
            'status' => 'required',
        ]);
        $user = User::findOrFail($request->user_id);
        $user->status = $request->status;
        $user->save();

        return response()->json(['success' => true, 'message' => 'User status updated successfully.']);
    }
    /**
     * Search for GiggyCode users based on filters.
     */
    public function searchGiggyCodeUsers(Request $request)
    {
        $query = User::query();

        // Apply search filters
        if ($request->filled('name')) {
            $query->where('full_name', 'like', '%' . $request->name . '%');
        }

        if ($request->filled('email')) {
            $query->where('email', 'like', '%' . $request->email . '%');
        }

        if ($request->filled('id')) {
            $query->where('id', $request->id);
        }

        // Fetch filtered users with certificate counts
        $users = $query->withCount(['certificates', 'certificates as verified_certificates' => function ($q) {
            $q->where('verification_status', 'Verified');
        }])->paginate(10);

        return view('users.view-giggycode', compact('users'));
    }
}
